import socket
from http.client import HTTPConnection
from xmlrpc.client import ServerProxy, Transport, ProtocolError
import sys
import inspect
import time
from pprint import pprint
import logging
import re

# create logger
log = logging.getLogger(__name__)
log.setLevel(logging.INFO)

# create console handler and set level to debug
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)

# create formatter
fmt = logging.Formatter('%(asctime)s - %(name)s - %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p')

# add formatter to ch
ch.setFormatter(fmt)

# add ch to logger
log.addHandler(ch)

TaskOpr = ['Clone', 'Clone', 'Clone', 'Clone', 'Clone',
		'Wipe', 'Wipe', 'Wipe', 'Wipe', 'Wipe',
		'Hash', 'Hash', 'Hash', 'Hash', 'Hash',
		'Push', 'Push', 'Push', 'Push', 'Push']

class TimeoutTransport(Transport):
	def __init__(self, timeout=socket._GLOBAL_DEFAULT_TIMEOUT, *args, **kwargs):
		Transport.__init__(self, *args, **kwargs)
		self.timeout = timeout

	def make_connection(self, host):
		if self._connection and host == self._connection[0]:
			return self._connection[1]

		chost, self._extra_headers, x509 = self.get_host_info(host)
		self._connection = host, HTTPConnection(chost, timeout=self.timeout)
		return self._connection[1]

class TimeoutServerProxy(ServerProxy):
	def __init__(self,uri,to=5):
		t=TimeoutTransport(timeout=to)
		ServerProxy.__init__(self,uri,transport=t, allow_none=True)

class XmlRpcApi(object):
	ROBOT_LIBRARY_SCOPE = 'GLOBAL'
	def __init__(self, host, port=80):
		self.host = host
		self.sp = TimeoutServerProxy('http://%s:%d' % (host, port), to=15)

		# Add all available xml-rpc methods as class methods pointing to __call
		try:
			l=self.sp.system.listMethods()
		except (socket.timeout, socket.gaierror, socket.error, socket.herror) as e:
			self.ConnectionLost = True
			raise RuntimeError("Connection to {0} Failed: {1}".format(self.host, e))
		except (ProtocolError) as e:
			self.ConnectionLost = True
			raise RuntimeError("{0} Not a valid Logicube unit".format(self.host))

		for m in l:
			setattr(self, m, self.__call)

		self.ConnectionLost = False

	def __call(self, args=None, v=0):
		# Get the method name used to invoke this function
		f=inspect.stack()[1][4][0].strip('\n').split('(')[0].split('.')
		n = f[len(f)-1]

		if v:
			log.debug("Calling method: {0}".format(n))
		m = getattr(self.sp, n)
		try:
			if args is None:
				r = m()
			else:
				r = m(args)
		except (socket.timeout, socket.gaierror, socket.error, socket.herror) as e:
			self.ConnectionLost = True
			log.error("[!] Connection to {0} Failed: {1}".format(self.host, e))
			raise RuntimeError("Connection to {0} Failed: {1}".format(self.host, e))

		if r['Method'] == n and r['Result'] == "COMMAND_COMPLETED":
			#if v:
			log.debug(r)
			return r
		else:
			log.debug(r)
			log.error("[!] Call '%s' failed! (%s)" % (n, r['Result']))
			raise RuntimeError("Call '%s' failed! (%s)" % (n, r['Result']))

	def call(self, method, args=None, v=0):
		m = getattr(self.sp, method)
		try:
			if args is None:
				r = m()
			else:
				r = m(args)
		except (socket.timeout, socket.gaierror, socket.error, socket.herror) as e:
			self.ConnectionLost = True
			log.error("Connection to {0} Failed: {1}".format(self.host, e))
			raise RuntimeError("Connection to {0} Failed: {1}".format(self.host, e))

		if r['Method'] == method and r['Result'] == "COMMAND_COMPLETED":
			if v:
				log.debug(r)
			return r
		else:
			log.debug(r)
			log.error("[!] Call '%s' failed! (%s)" % (n, r['Result']))
			raise RuntimeError("Call '%s' failed! (%s)" % (method, r['Result']))

	def get_unit_serial(self):
		self.ApplicationVersion()
		return s.ApplicationVersion()["SerialNumber"]

	def get_unit_platform(self):
		self.ApplicationVersion()
		return s.ApplicationVersion()["Platform"]

	def get_unit_identifier(self):
		self.ApplicationVersion()
		return s.ApplicationVersion()["HostName"]

	def get_all_bays(self, filter=None):
		l=self.BayStatus()
		s=[]
		if filter:
			regx = re.compile(filter)
			s=[b[0]['Bay'] for b in l['Drives'] if regx.match(b[0]['Bay']) is not None and b[0]['Connected']]
		else:
			s=[b[0]['Bay'] for b in l['Drives'] if b[0]['Connected']]
		pprint(l)
		pprint(s)
		return s

	def get_source_bays(self, filter=None):
		l=self.BayStatus()
		s=[]
		if filter:
			regx = re.compile(filter)
			s=[b[0]['Bay'] for b in l['Drives'] if b[0]['Role'] == 'Master' and regx.match(b[0]['Bay']) is not None and b[0]['Connected']]
		else:
			s=[b[0]['Bay'] for b in l['Drives'] if b[0]['Role'] == 'Master' and b[0]['Connected']]
		return s

	def get_destination_bays(self, filter=None):
		l=self.BayStatus()
		s=[]
		if filter:
			regx = re.compile(filter)
			s=[b[0]['Bay'] for b in l['Drives'] if 'Target' in b[0]['Role'] and regx.match(b[0]['Bay']) is not None and b[0]['Connected']]
		else:
			s=[b[0]['Bay'] for b in l['Drives'] if 'Target' in b[0]['Role'] and b[0]['Connected']]
		return s

	def get_source_drives(self):
		l=self.BayStatus()
		s=[b for b in l['Drives'] if b[0]['Role'] == 'Master' and b[0]['Connected']]
		return s

	def get_destination_drives(self):
		l=self.BayStatus()
		s=[b for b in l['Drives'] if 'Target' in b[0]['Role'] and b[0]['Connected']]
		return s

	def power_cycle_bay(self, bay):
		r=self.GetBayRole(bay)
		# get current role
		role = r['Role']
		
		# power off bay
		log.info("[*] Powering OFF bay {0} @ {1}".format(bay, self.host))
		self.AssignBayRole({'Bay': bay, 'Role': 'Not Available'})
		
		# wait for 5 seconds
		time.sleep(5)
		
		# reset the old bay-role
		log.info("[*] Powering ON bay {0} @ {1}".format(bay, self.host))
		self.AssignBayRole({'Bay': bay, 'Role': role})
		
		# wait for bay
		waiting = True
		tries = 10
		while waiting:
			time.sleep(3)
			r=self.BayStatus()
			drives=[b for b in r['Drives'] if b[0]['Role'] == role and b[0]['Connected']]
			for d in drives:
				if d[0]['Bay'] == bay:
					waiting = False
			tries -= 1
			if tries == 0:
				log.warning("[*] Bay {0} @ {1}: Reset Failed!!".format(bay, self.host))
				return False

		log.info("[*] Bay {0} @ {1} is Ready".format(bay, self.host))
		return True

	def is_repo_mounted(self, repo, fs=None):
		repos=self.RepositoryList()
		for r in repos['Repositories']:
			if r['Name'] == repo:
				if fs is not None:
					log.info("[*] Repo={0}, Type={1}, Encrypted={2}, Status={3}".format(repo, r['Type'], r['Encrypted'], r['Status']))
					return (r['Status'] == 'MOUNTED' and r['Type'] == fs)
				return (r['Status'] == 'MOUNTED')
		return False

	def get_destination_repos(self):
		res=self.RepositoryList()
		return [r for r in res['Repositories'] if r['Role'] in ['Target', 'Both'] and r['Status'] == 'MOUNTED']

	def get_source_repos(self):
		res=self.RepositoryList()
		return [r for r in res['Repositories'] if r['Role'] in ['Master', 'Both'] and r['Status'] == 'MOUNTED']

	def is_network_repo_mounted(self, path):
		repos=self.GetDestinationRepos()
		r = [r for r in repos if 'Path' in r and r['Path'] == path]
		return (len(r) > 0)

	def add_network_repo(self, repo={}):
		s=self.RepositoryAdd(repo)
		if s['Result'] != 'COMMAND_COMPLETED':
			return False, s['Result']

		retry = 5
		while retry:
			if self.IsNetworkRepoMounted(repo['Path']):
				return True, 'Mounted'
			retry -= 1
			time.sleep(3)

		return False, 'Timedout'

	def is_iscsi_portal_connected(self, ip):
		l=self.IscsiList()
		r = [p['Portal'].strip(':')[0] for p in l['Portals'] if ip in p['Portal'] and p['Connected']]
		return (len(r) > 0)

	def GenerateSourceDestinationPairs(self, sfilter, dfilter):
		pairs = {}
		srcs = self.get_source_bays()
		dsts = self.get_destination_bays()
		src = [s for s in srcs if re.match(sfilter, s) is not None]
		dst = [d for d in dsts if re.match(dfilter, d) is not None]
		while len(src) and len(dst):
			s = src.pop(0)
			if len(src) == 0:
				pairs[s] = dst
			else:
				pairs[s] = [dst.pop(0)]
		return pairs

	def get_sas_sata_pairs(self):
		return self.GenerateSourceDestinationPairs(r'^SAS_S', r'^SAS_D|SATA_D')

	def get_pcie_pairs(self):
		return self.GenerateSourceDestinationPairs(r'^PCIe_S', r'^PCIe_D')

	def get_usb_pairs(self):
		return self.GenerateSourceDestinationPairs(r'^USB_S', r'^USB_D')

	def get_tbt_pairs(self):
		return self.GenerateSourceDestinationPairs(r'^TBT_S', r'^TBT_D')

	def get_all_pairs(self):
		p = self.get_tbt_pairs()
		p.update(self.get_usb_pairs())
		p.update(self.get_pcie_pairs())
		p.update(self.get_sas_sata_pairs())
		return p

	def get_pairs(self):
		p = self.GenerateSourceDestinationPairs(r'.*', r'.*')
		pprint(p)
		return p

	def get_host_statistics(self):
		s = self.ApplicationVersion()
		pprint(s)
		return s

	def get_task_id(self, task='Image:1'):
		ids = {'Image':0, 'Wipe':5, 'Hash':10, 'Push':15}
		t = task.split(':')
		return ids[t[0]] + int(t[1])

	def task_add(self, task):
		t = self.get_task_id(task)
		op = task.split(':')[0]
		if op == 'Image':
			op = 'Clone'
		self.TaskReserve({'Operation': op, 'Task': t})

	def task_delete(self, task):
		t = self.get_task_id(task)
		self.TaskRelease(t)

	def task_delete_all(self, opr):
		for t in range(1, 6):
			self.task_delete('{0}:{1}'.format(opr, t))

	def task_set_targets(self, task, filter=None):
		t = self.get_task_id(task)
		tgt = []
		if isinstance(filter, list):
			filter = '|'.join(filter)
		if task.split(':')[0] == 'Hash':
			tgt = self.get_all_bays(filter)
		else:
			tgt = self.get_destination_bays(filter)
		self.TaskSetConfiguration({'Task': t, 'Targets': tgt})

	def task_set_master(self, task, filter=None):
		t = self.get_task_id(task)
		src = self.get_source_bays(filter)
		pprint(src)
		self.TaskSetConfiguration({'Task': t, 'Master': src[0]})

	def task_set_repositories(self, task, repos=None):
		t = self.get_task_id(task)
		if isinstance(repos, list):
			repos = '|'.join(repos)
		tgt = self.get_destination_bays(repos)
		cplist = ['/' for t in tgt]
		pprint("{0}, {1}".format(tgt, cplist))
		self.TaskSetConfiguration({'Task': t, 'RepositoryList': tgt, 'CapturePathList': cplist})

	def task_set_hash(self, task, hash='SHA-1', vfy=None):
		t = self.get_task_id(task)
		config={'Task': t, 'HashType': hash}
		if vfy is not None and isinstance(vfy, bool):
			config['VerifyHash'] = vfy
		else:
			config['VerifyHash'] = True
			config['VerifyCaseHash'] = vfy
		pprint(config)
		self.TaskSetConfiguration(config)

	def task_set_config(self, task=None, cfg=None):
		config=cfg
		if task:
			config['Task'] = self.get_task_id(task)
		pprint(config)
		self.TaskSetConfiguration(config)

	def task_run(self, task):
		t = self.get_task_id(task)
		r=self.TaskStart(t)
		#assert r['Status']
		pprint(r)

	def is_task_ready(self, task):
		t = self.get_task_id(task)
		l=self.TaskStatus(t)
		return (l['TaskState'] == 'TASK_RESERVED')

	def is_task_running(self, task):
		t = self.get_task_id(task)
		l=self.TaskStatus(t)
		return (l['TaskState'] == 'TASK_WORKING')

	def task_status_get(self, task):
		t = self.get_task_id(task)
		l=self.TaskStatus(t)
		return l['TaskState']

	def task_get_source_hash(self, task, src):
		t = self.get_task_id(task)
		l=self.TaskStatus(t)
		regx=re.compile(src)
		for h in l['ComputedHashValues']:
			for k in h.keys():
				pprint(k)
				if regx.match(k):
					return h[k]
		return None

	def task_get_expected_hash(self, task):
		t = self.get_task_id(task)
		l=self.TaskStatus(t)
		return l['ExpectedHashValue']

	def wait_for_task_completion(self, tasks):
		n = len(tasks)
		done = 0
		while done < n:
			done = 0
			time.sleep(2)
			for task in tasks:
				if self.is_task_running(task) is False:
					done += 1
				time.sleep(3)

